/*
  Align, A quick two sequence alignment Perl extension module in C.

	   Version 0.01, Copyright 2009 (c) Hui-Hsien Chou

  This code is free software; you can redistribute it and/or modify it
  under the terms of the GNU General Public License version 3 as
  published by the Free Software Foundation. This code is distributed in
  the hope that it will be useful, but WITHOUT ANY WARRANTY; without
  even the implied warranty of MERCHANTABILITY or FITNESS FOR A
  PARTICULAR PURPOSE. See the GNU General Public License v3 at
  <http://www.gnu.org/licenses/licenses.html> for more details.

  To obtain the latest version visit <http://www.complex.iastate.edu>
*/

#include "EXTERN.h"
#include "perl.h"
#include "XSUB.h"

#include "ppport.h"

#include <string.h>
#include <stdlib.h>

#define NIL 0
#define LFT 1
#define DIA 2
#define TOP 3

#define Score(x, y) splice_score_base[(y)*width+(x)]
#define Direc(x, y) splice_direc_base[(y)*width+(x)]

static int convert[128]={
  -1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,
  -1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,
  -1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,15,-1,-1,
  -1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,
/* @  A  B  C  D  E  F  G  H  I  J  K  L  M  N  O
   P  Q  R  S  T  U  V  W  X  Y  Z                */
  -1, 0,14, 1,11,-1,-1, 2,12,-1,-1,10,-1, 7,15,-1,
  -1,-1, 5, 9, 3, 4,13, 8,15, 6,-1,-1,-1,-1,-1,-1,
  -1, 0,14, 1,11,-1,-1, 2,12,-1,-1,10,-1, 7,15,-1,
  -1,-1, 5, 9, 3, 4,13, 8,15, 6,-1,-1,-1,-1,-1,-1};

static int match[16][16]={
/*       A   C   G   T   U   R   Y   M   W   S   K   D   H   V   B   N */
/*A*/ {  5,-14,-14,-14,-14,  2,-14,  2,  2,-14,-14,  1,  1,  1,-14,  0,},
/*C*/ {-14,  5,-14,-14,-14,-14,  2,  2,-14,  2,-14,-14,  1,  1,  1,  0,},
/*G*/ {-14,-14,  5,-14,-14,  2,-14,-14,-14,  2,  2,  1,-14,  1,  1,  0,},
/*T*/ {-14,-14,-14,  5,  5,-14,  2,-14,  2,-14,  2,  1,  1,-14,  1,  0,},
/*U*/ {-14,-14,-14,  5,  5,-14,  2,-14,  2,-14,  2,  1,  1,-14,  1,  0,},
/*R*/ {  2,-14,  2,-14,-14,  2,-14,  1,  1,  1,  1,  1,  0,  1,  0,  0,},
/*Y*/ {-14,  2,-14,  2,  2,-14,  2,  1,  1,  1,  1,  0,  1,  0,  1,  0,},
/*M*/ {  2,  2,-14,-14,-14,  1,  1,  2,  1,  1,-14,  0,  1,  1,  0,  0,},
/*W*/ {  2,-14,-14,  2,  2,  1,  1,  1,  2,-14,  1,  1,  1,  0,  0,  0,},
/*S*/ {-14,  2,  2,-14,-14,  1,  1,  1,-14,  2,  1,  0,  0,  1,  1,  0,},
/*K*/ {-14,-14,  2,  2,  2,  1,  1,-14,  1,  1,  2,  1,  0,  0,  1,  0,},
/*D*/ {  1,-14,  1,  1,  1,  1,  0,  0,  1,  0,  1,  1,  0,  0,  0,  0,},
/*H*/ {  1,  1,-14,  1,  1,  0,  1,  1,  1,  0,  0,  0,  1,  0,  0,  0,},
/*V*/ {  1,  1,  1,-14,-14,  1,  0,  1,  0,  1,  0,  0,  0,  1,  0,  0,},
/*B*/ {-14,  1,  1,  1,  1,  0,  1,  0,  0,  1,  1,  0,  0,  0,  1,  0,},
/*N*/ {  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,},
/*       A   C   G   T   U   R   Y   M   W   S   K   D   H   V   B   N */
};

#define GAPOPEN -17
#define GAPGROW -5

static int left_penalty[]={ GAPOPEN, GAPGROW, GAPOPEN, GAPOPEN };
static int  top_penalty[]={ GAPOPEN, GAPOPEN, GAPOPEN, GAPGROW };

static char *a = 0;
static char *b = 0;
static int a_width=0;
static int b_height=0;

static int  *splice_score_base = 0;
static char *splice_direc_base = 0;
static int base_size = 0;

MODULE = Align		PACKAGE = Align

void
Align(f_seq, s_seq, f_left, f_right, s_left, s_right)
  SV *f_seq
  SV *s_seq
  int f_left = NO_INIT
  int f_right= NO_INIT
  int s_left = NO_INIT
  int s_right= NO_INIT
 CODE:
  register int left, top, direc, score, i, j;
  int max_score, max_i, max_j;

  int width, height;
  char *first = SvPV(f_seq, width);
  char *second = SvPV(s_seq, height);

  if (width<2 || height<2) {
    f_left = 0;
    f_right = 0;
    s_left = 0;
    s_right = 0;
    return;
  }

  /* convert first string to the 16 IUPAC DNA code, enlarge array if needed. */
  if (width>a_width) {
    free(a);
    a = malloc(sizeof(char)*width);
    a_width = width;
  }
  for (i=0; i<width; i++)
    a[i] = convert[first[i]];

  /* convert second string to the 16 IUPAC DNA code, enlarge array if needed. */
  if (height>b_height) {
    free(b);
    b = malloc(sizeof(char)*height);
    b_height = height;
  }
  for (j=0; j<height; j++)
    b[j] = convert[second[j]];

  /* enlarge the alignment matrices if needed. */
  i = width*height;
  if (i>base_size) {
    free(splice_score_base);
    splice_score_base = malloc(sizeof(int)*i);
    free(splice_direc_base);
    splice_direc_base = malloc(sizeof(char)*i);
    base_size = i;
  }

  /* setting the upper-left-most cell */
  Score(0,0)=max_score=match[*a][*b];
  Direc(0,0)=NIL;
  max_i=max_j=0;

  /* setting the first row */
  for (i=1; i<width; i++) {
    direc=NIL;
    score=match[a[i]][*b];
    left=Score(i-1,0)+left_penalty[Direc(i-1,0)];
    if (score<left) {
      direc=LFT; score=left;
    }
    Direc(i,0)=direc;
    Score(i,0)=score;
    if (score>max_score) {
      max_score=score; max_i=i; max_j=0;
    }
  }

  /* setting the first column */
  for (j=1; j<height; j++) {
    direc=NIL;
    score=match[*a][b[j]];
    top=Score(0,j-1)+top_penalty[Direc(0,j-1)];
    if (score<top) {
      direc=TOP; score=top;
    }
    Direc(0,j)=direc;
    Score(0,j)=score;
    if (score>max_score) {
      max_score=score; max_i=0; max_j=j;
    }
  }

  /* running dynamic programming for internal nodes */
  for (j=1; j<height; j++)
    for (i=1; i<width; i++) {
      score=Score(i-1,j-1);
      if (score>=0) {
	score+=match[a[i]][b[j]];
	direc=DIA;
      } else {
	score=match[a[i]][b[j]];
	direc=NIL;
      }
      top=Score(i,j-1)+top_penalty[Direc(i,j-1)];
      left=Score(i-1,j)+left_penalty[Direc(i-1,j)];
      if (top>left) {
	if (score<top) {
	  direc=TOP; score=top;
	}
      } else if (left>top) {
	if (score<left) {
	  direc=LFT; score=left;
	}
      } else if (score<top) {
	if (i*height>=j*width) {
	  direc=LFT; score=left;
	} else {
	  direc=TOP; score=top;
	}
      }
      Direc(i,j)=direc;
      Score(i,j)=score;
      if (score>=max_score) {
	max_score=score; max_i=i; max_j=j;
      }
    }

  /* bracktrack to find the longest alignment path */
  i=max_i; j=max_j;
  while (direc=Direc(i,j))
    switch (direc) {
    case LFT:
      i--; break;
    case DIA:
      i--; j--; break;
    case TOP:
      j--; break;
    }

  /* return the match coordinate on the strings */
  f_left = i;
  f_right = max_i;
  s_left = j;
  s_right = max_j;
 OUTPUT:
  f_left
  f_right
  s_left
  s_right
